from abc import ABC, abstractmethod
import asyncio
from typing import Any, AsyncGenerator, Coroutine
from logging import getLogger
logger = getLogger(__name__)

class ExecuteDataStream:
    """
    Represents a stream of data with associated metadata.

    This class provides methods to manage a stream of data, including reading and writing data
    asynchronously. It uses an asyncio Queue to store data items and manages metadata such as
    the name of the stream and its operational status.

    Attributes:
    - __metadata: A dictionary containing metadata about the data stream.
    - __data_queue: An asyncio Queue for storing data items.

    表示了含有元数据的数据流。

    该类提供了用于管理数据流的方法，包括异步读写数据。它使用 asyncio 队列来存储数据项，
    并管理诸如数据流名称和操作状态等元数据。

    属性:
    - __metadata: 包含关于数据流元数据的字典。
    - __data_queue: 用于存储数据项的 asyncio 队列。
    """
    def __init__(self,
                 name: str = "ExecuteDataStream",
                 maxsize: int = 8):
        """
        Initialize the ExecuteDataStream with an empty data buffer and metadata dictionary.
        
        Parameters:
        - name: str, the name of the ExecuteDataStream, defaults to "ExecuteDataStream".
        - maxsize: int, the maximum size of the data queue, defaults to 8.

        初始化 ExecuteDataStream 类的一个实例，创建一个空的数据缓冲区和元数据字典。

        参数:
        - name: str, 数据流的名字，默认为 "ExecuteDataStream"。
        - maxsize: int, 数据队列的最大容量，默认为 8。
        """
        self.__metadata = {
            "__name__": name,
            "__running__": True,
            "__open_life__": True,
        }
        self.__data_queue = asyncio.Queue(maxsize)

    async def iter_read(self, timeout: float | None = None) -> AsyncGenerator:
        """
        Asynchronously iterate over data in the stream, yielding each item as it becomes available.

        Parameters:
        - timeout: float | None, the time to wait for data before timing out.

        Returns:
        - AsyncGenerator: An asynchronous generator that yields data items as they become available.

        异步迭代数据流中的数据，当数据可用时生成每个数据项（阻塞等待数据）。

        参数:
        - timeout: float | None, 等待数据的时间，超时则停止等待。

        返回:
        - AsyncGenerator: 当数据可用时生成数据项的异步生成器。
        """
        while True:
            res = await self.read(timeout)
            if res is None:
                break
            yield res

    async def read(self, timeout: float | None = None) -> Any | None:
        """
        Asynchronously read data from the data stream.

        Parameters:
        - timeout: float | None, the time to wait for data before timing out.

        Returns:
        - Any | None: The data item or None if no data is available within the timeout period.

        异步从数据流中读取数据。

        参数:
        - timeout: float | None, 等待数据的时间，超时则停止等待。

        返回:
        - Any | None: 数据项或 None ，如果在超时期间内没有数据可用。
        """
        if not self.__metadata.get('__running__', False):
            return None

        try:
            data = await asyncio.wait_for(self.__data_queue.get(), timeout)
            return data.get("data", None)
        except asyncio.TimeoutError:
            return None

    async def write(self,
                    data: Any,
                    timeout: float | None = None,
                    life_time: int = -1) -> bool:
        """
        Asynchronously write data to the data stream.

        #### Note: Please do not write None as it will be considered an error and thus discarded.

        Parameters:
        - data: Any, the data to be written.
        - timeout: float | None, the time to wait for writing before timing out.
        - life_time: int, the number of iterations the data should remain in the stream.

        Returns:
        - bool: True if the data was successfully written, False otherwise.

        异步向数据流中写入数据。
        
        #### 注意请不要写入 None 因为 None 会被认为是 Error ，所以会被抛弃 。

        参数:
        - data: Any, 要写入的数据。
        - timeout: float | None, 写入等待时间，超时则停止等待。
        - life_time: int, 数据在数据流中的存活周期数。

        返回:
        - bool: 如果数据成功写入则返回 True ，否则返回 False。
        """
        if not self.__metadata.get('__running__', False):
            return False
        try:
            if life_time == 0 or data is None:
                return False
            if self.__metadata.get('__open_life__', False):
                life_time -= 1
            await asyncio.wait_for(self.__data_queue.put({
                'data': data,
                'life': life_time,
            }), timeout)
            return True
        except asyncio.TimeoutError:
            return False


class ExecuteProcessor(ABC):
    """
    An abstract base class for processing data streams.

    This class represents a processor that can handle incoming data from one stream and send processed data to another stream.

    Usage:
    - Inherit from this class and implement the `execute` method.
    - Call the `set_stream` method to set `read_stream` and `write_stream`.
    - Call the `process` method to start the processor (if there are custom requirements, you can refactor this function, but the refactoring must comply with the interface standard).

    Attributes:
    - _read_stream: (Should not be used directly) An instance of `ExecuteDataStream` for reading data.
    - _write_stream: (Should not be used directly) An instance of `ExecuteDataStream` for writing data.

    一个抽象基类，用于处理数据流。

    该类表示一个处理器，可以从一个数据流读取数据，并将处理后的数据发送到另一个数据流。

    使用方法:
    - 继承该类并实现 `execute` 方法。
    - 调用 `set_stream` 方法设置 `read_stream` 和 `write_stream`。
    - 调用 `process` 方法启动处理器（如果有定制化要求可重构该函数，但重构必须符合接口标准）。

    属性:
    - _read_stream: （不应当直接使用）用于读取数据的 `ExecuteDataStream` 实例。
    - _write_stream: （不应当直接使用）用于写入数据的 `ExecuteDataStream` 实例。
    """
    def __init__(self):
        """
        Initialize the `ExecuteProcessor` with default read and write streams set to `None`.

        初始化 `ExecuteProcessor` 类的一个实例，将默认的读取和写入数据流设置为 `None`。
        """
        self._read_stream:  ExecuteDataStream | None = None
        self._write_stream: ExecuteDataStream | None = None

    def set_stream(self,
                   read_stream: ExecuteDataStream | None = None,
                   write_stream: ExecuteDataStream | None = None):
        """
        Set the read and write streams for this processor.

        #### Note: Once `read_stream` and `write_stream` are set, this instance should not accept new settings.

        Parameters:
        - read_stream: An instance of `ExecuteDataStream` for reading data.
        - write_stream: An instance of `ExecuteDataStream` for writing data.

        设置此处理器的读取和写入数据流。

        #### 注意一旦设置了 `read_stream` 和 `write_stream` ，则该实例将不应该接受新的设置。

        参数:
        - read_stream: 用于读取数据的 `ExecuteDataStream` 实例。
        - write_stream: 用于写入数据的 `ExecuteDataStream` 实例。
        """
        self._read_stream = read_stream or self._read_stream
        self._write_stream = write_stream or self._write_stream

    async def _iter_process(self):
        """
        Iterate over the data in the read stream and process it, then write the results to the write stream.

        #### Note: If the `process` method is refactored in the future, this function may serve as a scaffold.

        迭代读取数据流中的数据并对其进行处理，然后将结果写入写入数据流。

        #### 注意如果未来重构 `process` 方法，可能需要该函数作为脚手架。
        """
        if self._read_stream is None or self._write_stream is None:
            raise ValueError("read_stream or write_stream is None")
        try:
            async for data in self._read_stream.iter_read():
                result = await self.execute(data)
                if isinstance(result, AsyncGenerator):
                    async for res in result:
                        await self._write_stream.write(res)
                        await asyncio.sleep(0)
                else:
                    await self._write_stream.write(result)
                await asyncio.sleep(0)
        except Exception as e:
            logger.exception(f"An error occurred during processing: {e}")
            # await self._write_stream.write(e)

    @abstractmethod
    async def execute(self, data) -> AsyncGenerator[Any, Any] | Coroutine[Any, Any, Any]:
        """
        Process the given data.

        Parameters:
        - data: The data to be processed.

        Returns:
        - AsyncGenerator | Any: The processed data or an asynchronous generator of processed data.

        处理给定的数据。

        参数:
        - data: 要处理的数据。

        返回:
        - AsyncGenerator | Any: 处理后的数据或处理后数据的异步生成器。
        """
        pass

    async def process(self):
        """
        Future entry point for starting the processor.

        未来启动处理器的调用入口。
        """
        await self._iter_process()


class ExecutePipeline(ExecuteProcessor):
    """
    A class representing a pipeline of processors.

    This class manages a sequence of `ExecuteProcessor` instances and coordinates their execution.

    Attributes:
    - _processors: A list of `ExecuteProcessor` instances.
    - _streams: A list of `ExecuteDataStream` instances, including the first and last streams which can be set externally.
    - _tasks: A list of asyncio Tasks representing the execution of each processor.

    一个表示处理器管道的类。

    该类管理一系列 `ExecuteProcessor` 实例，并协调它们的执行。

    属性:
    - _processors: `ExecuteProcessor` 实例列表。
    - _streams: `ExecuteDataStream` 实例列表，包括可以外部设置的第一个和最后一个数据流。
    - _tasks: 表示每个处理器执行的 asyncio Task 列表。
    """

    def __init__(self, processors: list[ExecuteProcessor]):
        """
        Initialize the `ExecutePipeline` with a list of processors.

        Parameters:
        - processors: A list of `ExecuteProcessor` instances.

        使用处理器列表初始化 `ExecutePipeline`。

        参数:
        - processors: `ExecuteProcessor` 实例列表。
        """
        self._processors = processors
        self._streams = [None] + [ExecuteDataStream() for _ in range(len(self._processors) - 1)] + [None]
        self._tasks:list[asyncio.Task] = []
        self.__is_executed = False

    def set_stream(self,
                   read_stream: ExecuteDataStream | None = None,
                   write_stream: ExecuteDataStream | None = None):
        """
        Set the input and output streams for the pipeline.

        Parameters:
        - read_stream: The input stream for the pipeline.
        - write_stream: The output stream for the pipeline.

        设置管道的输入和输出数据流。

        参数:
        - read_stream: 管道的输入数据流。
        - write_stream: 管道的输出数据流。
        """
        self._streams[0] = write_stream or self._streams[0]
        self._streams[-1] = read_stream or self._streams[-1]

    def _get_stream(self, index: int) -> ExecuteDataStream:
        """
        Get the stream at the specified index.

        Parameters:
        - index: The index of the stream to retrieve.

        Returns:
        - ExecuteDataStream: The stream at the specified index.

        获取指定索引处的数据流。

        参数:
        - index: 要获取的数据流的索引。

        返回:
        - ExecuteDataStream: 指定索引处的数据流。
        """
        stream = self._streams[index]
        if stream is None:
            raise ValueError("Stream not found")
        return stream

    async def write(self,
                    data,
                    index = 0,
                    timeout = None,
                    life_time = -1):
        """
        Write data to the pipeline's input data stream.

        Parameters:
        - data: The data to write.
        - index: The index of the stream to write to. Default is 0 (the first stream).
        - timeout: The timeout for writing the data.
        - life_time: The life time of the data.

        将数据写入管道的输入数据流。

        参数:
        - data: 要写入的数据。
        - index: 写入数据的数据流索引，默认为 0 （第一个数据流）。
        - timeout: 写入数据的超时时间。
        - life_time: 数据的生命周期。
        """
        await self._get_stream(index).write(data, timeout, life_time)

    async def read(self, index: int = -1, timeout: float | None = None):
        """
        Read data from the last processor's data stream.

        Parameters:
        - index: The index of the stream to read from. Default is -1 (the last stream).
        - timeout: The timeout for reading the data.

        Returns:
        - Any: The data read from the stream.

        从最后一个处理器的数据流中读取数据。

        参数:
        - index: 读取数据的数据流索引，默认为 -1 （最后一个数据流）。
        - timeout: 读取数据的超时时间。

        返回:
        - Any: 从数据流中读取的数据。
        """
        return await self._get_stream(index).read(timeout)

    def iter_read(self, index: int = -1, timeout: float | None = None):
        """
        Iterate over the data in the specified stream.

        Parameters:
        - index: The index of the stream to iterate over. Default is -1 (the last stream).
        - timeout: The timeout for reading the data.

        Returns:
        - AsyncIterator: An asynchronous iterator over the data in the stream.

        迭代指定数据流中的数据。

        参数:
        - index: 迭代数据的数据流索引，默认为 -1 （最后一个数据流）。
        - timeout: 读取数据的超时时间。

        返回:
        - AsyncIterator: 对数据流中的数据进行迭代的异步迭代器。
        """
        return self._get_stream(index).iter_read(timeout)

    def cancel(self):
        """
        Cancel all running tasks in the pipeline.

        取消管道中所有正在运行的任务。
        """
        if self._tasks is None:
            return
        for task in self._tasks:
            task.cancel()

    def start(self):
        """
        Start the pipeline by executing the `execute` method.

        通过执行 `execute` 方法来启动管道。
        """
        self.set_stream(ExecuteDataStream(), ExecuteDataStream())
        self.execute()
        return self

    def execute(self):
        """
        Execute the pipeline by setting up the streams for each processor and creating tasks for their execution.

        通过为每个处理器设置数据流并为其执行创建任务来执行管道。
        """
        if self.__is_executed:
            return
        self.__is_executed = True
        for i, processor in enumerate(self._processors):
            processor.set_stream(self._get_stream(i), self._get_stream(i + 1))
            self._tasks.append(
                asyncio.create_task(processor.process())
            )

    async def process(self):
        """
        Execute the pipeline and wait for all tasks to complete.

        执行管道并等待所有任务完成。
        """
        self.execute()
        await asyncio.gather(*self._tasks)

# 示例处理器
class ExampleProcessor(ExecuteProcessor):

    async def execute(self, data):
        return data + " processed"

async def main():
    processors:list[ExecuteProcessor] = [ExampleProcessor()]
    pipeline = ExecutePipeline(processors)
    pipeline.start()

    input_data = "Hello, world!"
    await pipeline.write(input_data)
    print("Pipeline started.")
    async def fun():
        async for data in pipeline.iter_read():
            print(data)
    task = asyncio.create_task(fun())

    while True:
        data = input("input data: ")
        await pipeline.write(data)
        await asyncio.sleep(0)
        # print(await pipeline.read())

# 使用示例
if __name__ == "__main__":
    asyncio.run(main())