import os
from pathlib import Path

import re

from typing import Any, Callable, Dict, Optional, List

import yaml

class YamlEnvLoader(yaml.SafeLoader):
    def __init__(self, stream):
        super(YamlEnvLoader, self).__init__(stream)
        self.add_implicit_resolver('!env_var', re.compile(r'\$\{(.+?)\}'), None)

    @staticmethod
    def env_var_constructor(loader, node):
        value = loader.construct_scalar(node)
        match = re.match(r'\$\{(.+?)\}', value)
        if match:
            env_var = match.group(1)
            return os.environ.get(env_var, value)
        return value
YamlEnvLoader.add_constructor('!env_var', YamlEnvLoader.env_var_constructor)

class PluginsConfig:
    @staticmethod
    def load_config(path: Path) -> Dict[str, Any]:
        """Load YAML configuration from the given path."""
        if not path.exists():
            raise FileExistsError(f"{path} not exists")
        if not path.is_file():
            raise FileNotFoundError(f"Config file at {path} does not exist.")
        with open(path, 'r') as f:
            return yaml.load(f, Loader=YamlEnvLoader)

    @staticmethod
    def recursive_load(base_path: Path, config: Dict[str, Any] | Any):
        """Recursively load configurations when 'plugin' and 'path' are present."""
        if isinstance(config, dict) and 'plugin' in config and 'path' in config:
            plugin_path:Path = base_path / config.get('path', '')
            nested_config = PluginsConfig.load_config(plugin_path)
            config.update(nested_config[config['plugin']])
        plg_root_path = plugin_path.parent / nested_config.get('root_path', None)
        if plg_root_path is not None:
            plugin_path = plg_root_path / nested_config[config['plugin']]['path']
        return config, plg_root_path, plugin_path
    
    @staticmethod
    def get_engine_conf(engine_name, base_path: Path, config_file: Path):
        """
        获取引擎配置
        """
        conf = PluginsConfig.load_config(config_file)
        return  PluginsConfig.recursive_load(base_path, conf.get(engine_name))